import asyncio
from .lib.protocol import _SiriDBProtocol
from .lib.protocol import _SiriDBInfoProtocol
from .lib.connection import SiriDBConnection
from .lib.defaults import DEFAULT_CLIENT_PORT
from .lib.client import SiriDBClient, SiriDBAsyncConnection, SiriDBConn
from .lib.constants import SECOND
from .lib.constants import MICROSECOND
from .lib.constants import MILLISECOND
from .lib.constants import NANOSECOND

__all__ = [
    'async_connect',
    'async_server_info',
    'connect',
    'SiriDBClient',
    'SiriDBProtocol',
    'SiriDBConn',
    'SECOND',
    'MICROSECOND',
    'MILLISECOND',
    'NANOSECOND'
]


class SiriDBProtocol(_SiriDBProtocol):

    def on_connection_made(self):
        pass

    def on_authenticated(self):
        pass

    def on_connection_lost(self, exc):
        pass


def connect(username,
            password,
            dbname,
            host='127.0.0.1',
            port=DEFAULT_CLIENT_PORT,
            loop=None,
            timeout=10,
            protocol=SiriDBProtocol):
    """WARNING: Creates a new asyncio event loop if none is given."""
    return SiriDBConnection(
        username,
        password,
        dbname,
        host=host,
        port=port,
        loop=loop,
        timeout=timeout,
        protocol=protocol)


async def async_connect(username,
                        password,
                        dbname,
                        host='127.0.0.1',
                        port=DEFAULT_CLIENT_PORT,
                        loop=None,
                        timeout=10,
                        keepalive=False,
                        protocol=SiriDBProtocol):

    connection = SiriDBAsyncConnection()
    await connection.connect(
        username,
        password,
        dbname,
        host=host,
        port=port,
        loop=loop,
        timeout=timeout,
        keepalive=keepalive,
        protocol=protocol)

    return connection


async def async_server_info(host='127.0.0.1',
                            port=DEFAULT_CLIENT_PORT,
                            loop=None,
                            timeout=10):
    loop = loop or asyncio.get_running_loop()
    client = loop.create_connection(
        lambda: _SiriDBInfoProtocol(None, None, None),
        host=host,
        port=port)
    transport, protocol = \
        await asyncio.wait_for(client, timeout=timeout)

    result = await protocol.future
    transport.close()
    return result



'''SiriDB Client protocol

:copyright: 2022, Jeroen van der Heijden (Cesbit.com)
'''
import asyncio
import qpack
from . import protomap
from .datapackage import DataPackage
from .exceptions import InsertError
from .exceptions import QueryError
from .exceptions import ServerError
from .exceptions import PoolError
from .exceptions import AuthenticationError
from .exceptions import UserAuthError
from .logging import logger as logging


_MAP = (
    lambda data: b'',
    lambda data: qpack.packb(data),
    lambda data: data
)


def _packdata(tipe, data=None):
    assert tipe in protomap.MAP_REQ_DTYPE, \
        'No data type found for message type: {}'.format(tipe)
    return _MAP[protomap.MAP_REQ_DTYPE[tipe]](data)


class _SiriDBProtocol(asyncio.Protocol):

    _connected = False

    _MAP = {
        # SiriDB Client protocol success response types
        protomap.CPROTO_RES_QUERY: lambda f, d: f.set_result(d),
        protomap.CPROTO_RES_INSERT: lambda f, d: f.set_result(d),
        protomap.CPROTO_RES_ACK: lambda f, d: f.set_result(None),
        protomap.CPROTO_RES_AUTH_SUCCESS: lambda f, d: f.set_result(True),
        protomap.CPROTO_RES_INFO: lambda f, d: f.set_result(d),
        protomap.CPROTO_RES_FILE: lambda f, d: f.set_result(d),

        # SiriDB Client protocol error response types
        protomap.CPROTO_ERR_MSG: lambda f, d: f.set_exception(
            RuntimeError(d.get('error_msg', None))),
        protomap.CPROTO_ERR_QUERY: lambda f, d: f.set_exception(
            QueryError(d.get('error_msg', None))),
        protomap.CPROTO_ERR_INSERT: lambda f, d: f.set_exception(
            InsertError(d.get('error_msg', None))),
        protomap.CPROTO_ERR_SERVER: lambda f, d: f.set_exception(
            ServerError(d.get('error_msg', None))),
        protomap.CPROTO_ERR_POOL: lambda f, d: f.set_exception(
            PoolError(d.get('error_msg', None))),
        protomap.CPROTO_ERR_USER_ACCESS: lambda f, d: f.set_exception(
            UserAuthError(d.get('error_msg', None))),
        protomap.CPROTO_ERR: lambda f, d: f.set_exception(
            RuntimeError(
                'Unexpected error occurred, view siridb log '
                'for more info')),
        protomap.CPROTO_ERR_NOT_AUTHENTICATED: lambda f, d: f.set_exception(
            AuthenticationError('This connection is not authenticated')),
        protomap.CPROTO_ERR_AUTH_CREDENTIALS: lambda f, d: f.set_exception(
            AuthenticationError('Invalid credentials')),
        protomap.CPROTO_ERR_AUTH_UNKNOWN_DB: lambda f, d: f.set_exception(
            AuthenticationError('Unknown database')),
        protomap.CPROTO_ERR_LOADING_DB: lambda f, d: f.set_exception(
            RuntimeError(
                'Error loading database, '
                'please check the SiriDB log files')),
        protomap.CPROTO_ERR_FILE: lambda f, d: f.set_exception(
            RuntimeError('Error retreiving file')),
    }

    def __init__(self, username, password, dbname):
        self._buffered_data = bytearray()
        self._data_package = None
        self._pid = 0
        self._requests = {}
        self._username = username
        self._password = password
        self._dbname = dbname
        self.auth_future = None

    def connection_made(self, transport):
        '''
        override asyncio.Protocol
        '''

        self._connected = True
        self.transport = transport

        self.remote_ip, self.port = transport.get_extra_info('peername')[:2]

        logging.debug(
            'Connection made (address: {} port: {})'
            .format(self.remote_ip, self.port))

        self.auth_future = self.send_package(protomap.CPROTO_REQ_AUTH,
                                             data=(self._username,
                                                   self._password,
                                                   self._dbname),
                                             timeout=10)

        self._password = None
        self.on_connection_made()

    def connection_lost(self, exc):
        '''
        override asyncio.Protocol
        '''
        self._connected = False

        logging.debug(
            'Connection lost (address: {} port: {})'
            .format(self.remote_ip, self.port))

        for pid, (future, task) in self._requests.items():
            task.cancel()
            if future.cancelled():
                continue
            future.set_exception(ConnectionError(
                'Connection is lost before we had an answer on package id: {}.'
                .format(pid)))

        self.on_connection_lost(exc)

    def data_received(self, data):
        '''
        override asyncio.Protocol
        '''
        self._buffered_data.extend(data)
        while self._buffered_data:
            size = len(self._buffered_data)
            if self._data_package is None:
                if size < DataPackage.struct_datapackage.size:
                    return None
                self._data_package = DataPackage(self._buffered_data)
            if size < self._data_package.length:
                return None
            try:
                self._data_package.extract_data_from(self._buffered_data)
            except KeyError as e:
                logging.error('Unsupported package received: {}'.format(e))
            except Exception as e:
                logging.exception(e)
                # empty the byte-array to recover from this error
                self._buffered_data.clear()
            else:
                self._on_package_received()
            self._data_package = None

    def send_package(self, tipe, data=None, is_binary=False, timeout=3600):
        self._pid += 1
        self._pid %= 65536  # pid is handled as uint16_t

        if not is_binary:
            data = _packdata(tipe, data)

        header = DataPackage.struct_datapackage.pack(
            len(data),
            self._pid,
            tipe,
            tipe ^ 255)

        self.transport.write(header + data)

        task = asyncio.ensure_future(self._timeout_request(self._pid,
                                                           timeout,
                                                           tipe))
        future = asyncio.Future()
        self._requests[self._pid] = (future, task)
        return future

    def on_connection_made(self):
        '''
        Called when a connection is made.
        The connection might not be useful until authentication has finished.
        '''
        pass

    def on_authenticated(self):
        '''
        Called when the connection is authenticated.
        '''
        pass

    def on_connection_lost(self, exc):
        '''
        Called when the connection is lost or closed.

        The argument is an exception object or None (the latter
        meaning a regular EOF is received or the connection was
        aborted or closed).
        '''
        pass

    async def _timeout_request(self, pid, timeout, tipe):
        await asyncio.sleep(timeout)
        if not self._requests[pid][0].cancelled():
            self._requests[pid][0].set_exception(TimeoutError(
                'Request timed out on PID {} ({})'
                .format(pid, protomap.TEXT_REQ_MAP.get(tipe, 'UNKNOWN'))))
        del self._requests[pid]

    def _on_package_received(self):
        try:
            future, task = self._requests.pop(self._data_package.pid)
        except KeyError:
            logging.error(
                'Package ID not found: {} ({})'.format(
                    self._data_package.pid,
                    protomap.TEXT_RES_MAP.get(
                        self._data_package.tipe,
                        'UNKNOWN')))
            return None

        # cancel the timeout task
        task.cancel()

        if future.cancelled():
            return

        self._MAP.get(self._data_package.tipe, lambda f, d: f.set_exception(
            TypeError(
                'Client received an unknown package type: {}'
                .format(self._data_package.tipe))))(
                    future,
                    self._data_package.data)


class _SiriDBInfoProtocol(_SiriDBProtocol):

    def connection_made(self, transport):
        '''
        override _SiriDBProtocol
        '''

        self.transport = transport
        self.remote_ip, self.port = transport.get_extra_info('peername')[:2]

        logging.debug(
            'Connection made (address: {} port: {})'
            .format(self.remote_ip, self.port))

        self.future = self.send_package(
                protomap.CPROTO_REQ_INFO,
                data=None,
                timeout=10)


class _SiriDBConnProtocol(_SiriDBProtocol):

    def on_connection_lost(self, exc):
        if hasattr(self, 'close_future'):
            self.close_future.set_result(None)
            delattr(self, 'close_future')



'''Protocol mapping (SiriDB version 2.0.x)

:copyright: 2022, Jeroen van der Heijden (Cesbit.com)
'''

# data types
DTYPE_NONE = 0
DTYPE_QPACK = 1
DTYPE_FILE = 2

# SiriDB Client protocol client request types
CPROTO_REQ_QUERY = 0
CPROTO_REQ_INSERT = 1
CPROTO_REQ_AUTH = 2
CPROTO_REQ_PING = 3
CPROTO_REQ_INFO = 4
CPROTO_REQ_LOADDB = 5
CPROTO_REQ_REGISTER_SERVER = 6
CPROTO_REQ_FILE_SERVERS = 7
CPROTO_REQ_FILE_USERS = 8
CPROTO_REQ_FILE_GROUPS = 9

MAP_REQ_DTYPE = {
    # SiriDB Client protocol client request types
    CPROTO_REQ_QUERY: DTYPE_QPACK,
    CPROTO_REQ_INSERT: DTYPE_QPACK,
    CPROTO_REQ_AUTH: DTYPE_QPACK,
    CPROTO_REQ_PING: DTYPE_NONE,
    CPROTO_REQ_INFO: DTYPE_NONE,
    CPROTO_REQ_LOADDB: DTYPE_QPACK,
    CPROTO_REQ_REGISTER_SERVER: DTYPE_QPACK,
    CPROTO_REQ_FILE_SERVERS: DTYPE_NONE,
    CPROTO_REQ_FILE_USERS: DTYPE_NONE,
    CPROTO_REQ_FILE_GROUPS: DTYPE_NONE
}

TEXT_REQ_MAP = {
    CPROTO_REQ_QUERY: 'CPROTO_REQ_QUERY',
    CPROTO_REQ_INSERT: 'CPROTO_REQ_INSERT',
    CPROTO_REQ_AUTH: 'CPROTO_REQ_AUTH',
    CPROTO_REQ_PING: 'CPROTO_REQ_PING',
    CPROTO_REQ_INFO: 'CPROTO_REQ_INFO',
    CPROTO_REQ_LOADDB: 'CPROTO_REQ_LOADDB',
    CPROTO_REQ_REGISTER_SERVER: 'CPROTO_REQ_REGISTER_SERVER',
    CPROTO_REQ_FILE_SERVERS: 'CPROTO_REQ_FILE_SERVERS',
    CPROTO_REQ_FILE_USERS: 'CPROTO_REQ_FILE_USERS',
    CPROTO_REQ_FILE_GROUPS: 'CPROTO_REQ_FILE_GROUPS'
}

# SiriDB Client protocol success server response types
CPROTO_RES_QUERY = 0
CPROTO_RES_INSERT = 1
CPROTO_RES_AUTH_SUCCESS = 2
CPROTO_RES_ACK = 3
CPROTO_RES_INFO = 4
CPROTO_RES_FILE = 5

# SiriDB Client protocol error server response types
CPROTO_ERR_MSG = 64
CPROTO_ERR_QUERY = 65
CPROTO_ERR_INSERT = 66
CPROTO_ERR_SERVER = 67
CPROTO_ERR_POOL = 68
CPROTO_ERR_USER_ACCESS = 69
CPROTO_ERR = 70
CPROTO_ERR_NOT_AUTHENTICATED = 71
CPROTO_ERR_AUTH_CREDENTIALS = 72
CPROTO_ERR_AUTH_UNKNOWN_DB = 73
CPROTO_ERR_LOADING_DB = 74
CPROTO_ERR_FILE = 75

# mapping between package types and data types
MAP_RES_DTYPE = {
    # SiriDB Client protocol success server response types
    CPROTO_RES_QUERY: DTYPE_QPACK,
    CPROTO_RES_INSERT: DTYPE_QPACK,
    CPROTO_RES_AUTH_SUCCESS: DTYPE_NONE,
    CPROTO_RES_ACK: DTYPE_NONE,
    CPROTO_RES_INFO: DTYPE_QPACK,
    CPROTO_RES_FILE: DTYPE_FILE,

    # SiriDB Client protocol error server response types
    CPROTO_ERR_MSG: DTYPE_QPACK,
    CPROTO_ERR_QUERY: DTYPE_QPACK,
    CPROTO_ERR_INSERT: DTYPE_QPACK,
    CPROTO_ERR_SERVER: DTYPE_QPACK,
    CPROTO_ERR_POOL: DTYPE_QPACK,
    CPROTO_ERR_USER_ACCESS: DTYPE_QPACK,
    CPROTO_ERR: DTYPE_NONE,
    CPROTO_ERR_NOT_AUTHENTICATED: DTYPE_NONE,
    CPROTO_ERR_AUTH_CREDENTIALS: DTYPE_NONE,
    CPROTO_ERR_AUTH_UNKNOWN_DB: DTYPE_NONE,
    CPROTO_ERR_LOADING_DB: DTYPE_NONE,
    CPROTO_ERR_FILE: DTYPE_NONE
}

TEXT_RES_MAP = {
    # SiriDB Client protocol success response types
    CPROTO_RES_QUERY: 'CPROTO_RES_QUERY',
    CPROTO_RES_INSERT: 'CPROTO_RES_INSERT',
    CPROTO_RES_AUTH_SUCCESS: 'CPROTO_RES_AUTH_SUCCESS',
    CPROTO_RES_ACK: 'CPROTO_RES_ACK',
    CPROTO_RES_INFO: 'CPROTO_RES_INFO',
    CPROTO_RES_FILE: 'CPROTO_RES_FILE',

    # SiriDB Client protocol error response types
    CPROTO_ERR_MSG: 'CPROTO_ERR_MSG',
    CPROTO_ERR_QUERY: 'CPROTO_ERR_QUERY',
    CPROTO_ERR_INSERT: 'CPROTO_ERR_INSERT',
    CPROTO_ERR_SERVER: 'CPROTO_ERR_SERVER',
    CPROTO_ERR_POOL: 'CPROTO_ERR_POOL',
    CPROTO_ERR_USER_ACCESS: 'CPROTO_ERR_USER_ACCESS',
    CPROTO_ERR: 'CPROTO_ERR',
    CPROTO_ERR_NOT_AUTHENTICATED: 'CPROTO_ERR_NOT_AUTHENTICATED',
    CPROTO_ERR_AUTH_CREDENTIALS: 'CPROTO_ERR_AUTH_CREDENTIALS',
    CPROTO_ERR_AUTH_UNKNOWN_DB: 'CPROTO_ERR_AUTH_UNKNOWN_DB',
    CPROTO_ERR_LOADING_DB: 'CPROTO_ERR_LOADING_DB',
    CPROTO_ERR_FILE: 'CPROTO_ERR_FILE'
}

# mapping used to find a file name for a package type
FILE_MAP = {
    'servers.dat': CPROTO_REQ_FILE_SERVERS,
    'users.dat': CPROTO_REQ_FILE_USERS,
    'groups.dat': CPROTO_REQ_FILE_GROUPS
}



'''SiriDB Client

SiriDB Client for python => 3.5 using asyncio.

:copyright: 2022, Jeroen van der Heijden (Cesbit.com)
'''
import asyncio
import functools
import random
from .protocol import _SiriDBProtocol, _SiriDBConnProtocol
from .connection import SiriDBAsyncConnection
from .exceptions import ServerError
from .exceptions import PoolError
from .constants import SECOND
from .constants import MICROSECOND
from .constants import MILLISECOND
from .constants import NANOSECOND
from .protomap import CPROTO_REQ_QUERY
from .protomap import CPROTO_REQ_INSERT
from .protomap import CPROTO_REQ_PING
from .logging import logger as logging


class _SiriDBClientProtocol(_SiriDBProtocol):

    _is_available = False

    def __init__(self, *args, trigger_connect, inactive_time):
        super().__init__(*args)
        self._trigger_connect = trigger_connect
        self._inactive_time = inactive_time

    def on_authenticated(self):
        self._is_available = True

    def on_connection_lost(self, exc):
        self._is_available = False
        self._trigger_connect()

    def set_available(self):
        if self._connected:
            self._is_available = True

    def set_not_available(self, loop):
        if self._is_available:
            self._is_available = False
            loop.call_later(self._inactive_time, self.set_available)


# never wait more than x seconds before trying to connect again
DEFAULT_MAX_WAIT_RETRY = 90

# default timeout used while connecting to a SiriDB server
DEFAULT_CONNECT_TIMEOUT = 10

# when a SiriDB server is marked as inactive, wait x seconds before releasing
# the inactive status.
DEFAULT_INACTIVE_TIME = 30


class SiriDBClient:
    '''
        Exception handling:

        - InsertError (can only be raised when using the insert() method)
            Make sure the data is correct because this only happens when
            SiriDB could not process the request. Its likely to fail again
            on a retry.
        - QueryError (can only be raised when using the query() method)
            Make sure the query is correct because this only happens when
            SiriDB could not process the query. Its likely to fail again.
        - PoolError
            SiriDB has no online server for at least one required pool
            Try again later after some reasonable delay.
        - AuthenticationError
            Raised when credentials are invalid or insufficient
        - IndexError
            Raised when the database does not exist (anymore)
        - TypeError
            Raised when an unknown package is received. (might be caused
            by running a different SiriDB version)
        - RuntimeError
            Raised when a general error message is received. This should not
            happen unless a new bug is discovered.
        - OverflowError (can only be raised when using the insert() method)
            Raise when integer values cannot not be packed due to an overflow
            error. (integer values should be signed and not more than 63 bits)
        - UserAuthError
            The user as no rights to perform the insert or query.
    '''

    def __init__(self,
                 username,
                 password,
                 dbname,
                 hostlist,
                 loop=None,
                 keepalive=True,
                 timeout=DEFAULT_CONNECT_TIMEOUT,
                 inactive_time=DEFAULT_INACTIVE_TIME,
                 max_wait_retry=DEFAULT_MAX_WAIT_RETRY):
        '''Initialize.
        Arguments:
            username: User with permissions to use the database.
            password: Password for the given username.
            dbname: Name of the database.
            hostlist: List with SiriDB servers. (all servers or a subset of
                      servers can be in this list.)

                      Example:
                      [
                          ('server1.local', 9000, {'weight': 3}),
                          ('server2.local', 9000),
                          ('backup1.local', 9000, {'backup': True})
                      ]

                      Each server should at least has a hostname and port
                      number. Optionally you can provide a dictionary with
                      extra options.

                      Available Options:
                      - weight : Should be a value between 1 and 9. A higher
                                 value gives the server more weight so it will
                                 be more likely chosen. (default 1)
                      - backup : Should be either True or False. When True the
                                 server will be marked as backup server and
                                 will only be chosen if no other server is
                                 available. (default: False)

        Keyword arguments:
            loop: Asyncio loop. When None the default event loop will be used.
            keepalive: SiriDB Version >= 0.9.35 supporting keep-alive packages
            timeout: Timeout used when reconnecting to a SiriDB server.
            inactive_time: When a server is temporary not available, for
                           example the server could be paused, we mark the
                           server inactive for x seconds.
            max_wait_retry: When the reconnect loop starts, we try to reconnect
                            in a seconds, then 2 seconds, 4, 8 and so on until
                            max_wait_retry is reached and then use this value
                            to retry again.
        '''
        self._username = username
        self._password = password
        self._dbname = dbname
        self._connection_pool = []
        self._keepalive = keepalive
        for host, port, *config in hostlist:
            config = config.pop() if config else {}
            client = SiriDBAsyncConnection()
            client.host = host
            client.port = port
            client.is_backup = config.get('backup', False)
            client.weight = config.get('weight', 1)
            assert 0 < client.weight < 10, \
                'weight should be value between 1 and 9'
            for _ in range(client.weight):
                self._connection_pool.append(client)
        self._connections = set(self._connection_pool)
        self._loop = loop or asyncio.get_running_loop()
        self._timeout = timeout
        self._connect_task = None
        self._max_wait_retry = max_wait_retry
        self._protocol = \
            functools.partial(_SiriDBClientProtocol,
                              trigger_connect=self._trigger_connect,
                              inactive_time=inactive_time)

    @property
    def is_closed(self):
        '''Can be used to check if close() has been called.'''
        return not self._retry_connect

    @property
    def connected(self):
        '''Can be used to check the client has any active connections'''
        return any(connection.connected for connection in self._connections)

    @staticmethod
    def _log_connect_result(result):
        for r in result:
            if r:
                msg = str(r) or type(r).__name__
                logging.error(msg)

    async def connect(self, timeout=None):
        self._retry_connect = True
        result = await self._connect(timeout)
        if result and set(result) - {None} and self._connect_task is None:
            self._connect_task = asyncio.ensure_future(self._connect_loop())
        return result

    def close(self):
        self._retry_connect = False
        if self._connect_task is not None:
            self._connect_task.cancel()
            self._connect_task = None
        for connection in self._connections:
            if connection.connected:
                connection.close()

    async def insert(self, data, timeout=300):
        end = self._loop.time() + timeout
        while True:
            connection = self._get_random_connection()

            try:
                result = await connection.insert(data, timeout)
            except (ConnectionError, ServerError) as e:
                logging.debug('Insert failed with error {!r}, trying another '
                              'server if one is available...'.format(e))
                if connection._protocol:
                    connection._protocol.set_not_available(self._loop)
            except PoolError as e:
                if self._loop.time() > end:
                    raise
                msg = str(e) or type(e).__name__
                logging.debug(msg)
                await asyncio.sleep(2)
            else:
                return result

    async def query(self, query, time_precision=None, timeout=60):
        assert isinstance(query, (str, bytes)), \
            'query should be of type str, unicode or bytes'

        assert time_precision is None or isinstance(time_precision, int), \
            'time_precision should be None or an int type.'

        end = self._loop.time() + timeout
        try_unavailable = True
        while True:
            connection = self._get_random_connection(try_unavailable)
            try:
                result = await connection.query(query,
                                                time_precision=time_precision,
                                                timeout=timeout)
            except (ConnectionError, ServerError) as e:
                logging.debug('Query failed with error {!r}, trying another '
                              'server if one is available...'.format(e))
                if connection._protocol:
                    connection._protocol.set_not_available(self._loop)
            except PoolError as e:
                if self._loop.time() > end:
                    raise
                msg = str(e) or type(e).__name__
                logging.debug(msg)
                await asyncio.sleep(2)
            else:
                return result

            # only try unavailable once
            try_unavailable = False

    async def _connect(self, timeout=None):  # the one that actually connects
        tasks = [
            connection.connect(
                self._username,
                self._password,
                self._dbname,
                host=connection.host,
                port=connection.port,
                loop=self._loop,
                keepalive=self._keepalive,
                timeout=timeout or self._timeout,
                protocol=self._protocol)
            for connection in self._connections
            if not connection.connected]
        if not tasks:
            return
        logging.debug('Trying to connect to {} servers...'
                      .format(len(tasks)))
        result = await asyncio.gather(*tasks, return_exceptions=True)
        self._log_connect_result(result)
        return result

    async def _connect_loop(self):  # the one that looks for connections
        sleep = 1
        try:
            while [connection
                   for connection in self._connections
                   if not connection.connected]:
                logging.debug('Reconnecting in {} seconds...'.format(sleep))
                await asyncio.sleep(sleep)
                if self._connect_task is None:
                    break
                await self._connect()
                if self._connect_task is None:
                    break
                sleep = min(sleep * 2, self._max_wait_retry)
        except asyncio.CancelledError:
            pass
        finally:
            self._connect_task = None

    def _trigger_connect(self):
        if self._retry_connect and self._connect_task is None:
            self._connect_task = asyncio.ensure_future(self._connect_loop())

    def _get_random_connection(self, try_unavailable=False):
        available = \
            [connection
             for connection in self._connection_pool
             if connection._protocol and connection._protocol._is_available]

        non_backups = \
            [connection
             for connection in available
             if not connection.is_backup]

        if non_backups:
            return random.choice(non_backups)

        if available:
            return random.choice(available)

        if try_unavailable:

            connections = \
                [connection
                 for connection in self._connection_pool
                 if connection.connected]

            if connections:
                return random.choice(connections)

        raise PoolError('No available connections found')


class SiriDBConn:

    MAX_RECONNECT_WAIT_TIME = 60
    MAX_RECONNECT_TIMEOUT = 10
    MAX_WRITE_RETRY = 120
    RECONNECT_ATTEMPT = 3

    def __init__(self,
                 username,
                 password,
                 dbname,
                 server,
                 port=9000,
                 loop=None):
        self._username = username
        self._password = password
        self._dbname = dbname
        self._server = server
        self._port = port
        self._loop = loop or asyncio.get_running_loop()
        self._reconnecting = False
        self._protocol = None

    async def _connect(self, timeout):
        client = self._loop.create_connection(
            lambda: _SiriDBConnProtocol(
                self._username,
                self._password,
                self._dbname),
            host=self._server,
            port=self._port)
        _transport, self._protocol = \
            await asyncio.wait_for(client, timeout=timeout)

        try:
            _res = await asyncio.wait_for(
                self._protocol.auth_future,
                timeout=timeout)
        except Exception as exc:
            _transport.close()
            raise exc

    async def _reconnect_loop(self):
        try:
            wait_time = 1
            timeout = 2
            protocol = self._protocol
            while True:
                host, port = self._server, self._port
                try:
                    await self._connect(timeout=timeout)
                except Exception as e:
                    logging.error(
                        f'Connecting to {host}:{port} failed: '
                        f'{e}({e.__class__.__name__}), '
                        f'Try next connect in {wait_time} seconds'
                    )
                else:
                    if protocol and protocol._connected:
                        # make sure the `old` connection will be dropped
                        self._loop.call_later(10.0, protocol.transport.close)
                    break

                await asyncio.sleep(wait_time)
                wait_time *= 2
                wait_time = min(wait_time, self.MAX_RECONNECT_WAIT_TIME)
                timeout = min(timeout+1, self.MAX_RECONNECT_TIMEOUT)
        finally:
            self._reconnecting = False

    def _reconnect(self):
        if self._reconnecting:
            return asyncio.sleep(1)
        self._reconnecting = True
        return self._reconnect_loop()

    def is_connected(self):
        return self._protocol and self._protocol._connected

    async def insert(self, data, timeout=300):
        result = await self._ensure_write(
            CPROTO_REQ_INSERT,
            data=data,
            timeout=timeout)
        return result

    def close(self):
        if self.is_connected():
            if not hasattr(self._protocol, 'close_future'):
                self._protocol.close_future = self._loop.create_future()
            self._protocol.transport.close()

    async def wait_closed(self):
        if self._protocol and hasattr(self._protocol, 'close_future'):
            await self._protocol.close_future

    async def connect(self, timeout=120):
        result = await self._ensure_write(CPROTO_REQ_PING, timeout=timeout)
        return result

    async def query(self, query, time_precision=None, timeout=60):
        assert isinstance(query, (str, bytes)), \
            'query should be of type str, unicode or bytes'
        assert time_precision in (
            None,
            SECOND,
            MICROSECOND,
            MILLISECOND,
            NANOSECOND), 'time_precision must be either None, 0, 1, 2, 3'
        result = await self._ensure_write(
            CPROTO_REQ_QUERY,
            data=(query, time_precision),
            timeout=timeout)
        return result

    async def _ensure_write(
            self,
            tipe, data=None, is_binary=False, timeout=None):
        retry = 0
        while True:
            retry += 1

            if not self.is_connected():
                if retry > self.MAX_WRITE_RETRY:
                    raise ConnectionError("Failed to create a connection")
                if retry == 1:
                    logging.info('Wait for a connection')
                await self._reconnect()  # ensure the re-connect loop
                continue

            try:
                res = await self._protocol.send_package(
                    tipe, data, is_binary, timeout)
            except (ServerError,
                    PoolError,
                    OSError,
                    asyncio.TimeoutError) as e:
                if retry > self.MAX_WRITE_RETRY:
                    raise e
                if retry % self.RECONNECT_ATTEMPT == 0:
                    await self._reconnect()
                else:
                    await asyncio.sleep(1.0)
                continue

            return res
